#include <vga.h>
#include <vgagl.h>
#include "conf.h"

#include "globdef.h"
#include "uidef.h"
#include "sigdef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "screendef.h"

#define RX_POWTIM_TRIGGER 340
#define RX_POWTIM_SHOW 350
#define POWTIM_TRIGMODES 4

float powtim_powfac;
float powtim_gain;
float powtim_fgain;
int powtim_xstep;
int powtim_trigmode;
int powtim_displaymode;
unsigned int powtim_y0;
unsigned int powtim_y1;
unsigned int powtim_y2;
int powtim_scalepix;
int pause_flag;

void make_rf_envelopes(void)
{
int i, j, mm, pa, pc;
float *zi;
float t1,t2,t3;
mm=twice_rxchan;
zi=&fft1_float[fft1_px];
// The backwards transformation of fft1 uses 16 bit integers
// and produces two time functions for each channel, one for
// strong signals and another one for weak signals.
// We need floating point to see keying waveforms on a log
// scale so here the backwards transform is computed
// with floating point numbers.
for(j=0; j<mm; j+=2)
  {
  for(i=fft1_first_point; i<=fft1_last_point; i++)
    {
    fft1_tmp[2*i  ]=zi[mm*i+j  ];
    fft1_tmp[2*i+1]=zi[mm*i+j+1];
    }
  for(i=0; i<fft1_first_point; i++)
    {
    fft1_tmp[2*i  ]=0;
    fft1_tmp[2*i+1]=0;
    }
  for(i=fft1_last_point+1; i<fft1_size; i++)
    {
    fft1_tmp[2*i  ]=0;
    fft1_tmp[2*i+1]=0;
    }    
  if( (ui.input_mode&IQ_DATA) != 0)fft_iqshift(fft1_size, fft1_tmp);
  fftback(fft1_size, fft1_n, fft1_tmp, fft2_tab, fft2_permute);
  pa=timf2_pa;
  pc=(i+timf2_pa+timf2_mask)&timf2_mask;
  for(i=0; i<fft1_size/2; i++)
    {    
    t1=fft2_float[mm*pa+j  ]+fft1_tmp[2*i  ];
    t2=fft2_float[mm*pa+j+1]+fft1_tmp[2*i+1];
    fft2_float[mm*pa+j  ]=powtim_powfac*sqrt(t1*t1+t2*t2);
    if(fft2_float[mm*(i+timf2_pa)+j  ] !=0)
      {
      t3=atan2(t2,t1);
      fft2_float[mm*pa+j+1]=t3;
      t3-=fft2_float[mm*pc+j+1];
      if(t3 > PI_L)t3-=2*PI_L;
      if(t3 <-PI_L)t3+=2*PI_L;
      fft2_float[mm*pc+j+1]=t3;
      }
    else
      {
      fft2_float[mm*pa+j+1]=0;
      }  
    pc=pa;
    pa=(pa+1)&timf2_mask;
    }
  pa=(timf2_pa+fft1_size/2+timf2_mask)&timf2_mask;
  for(i=0; i<fft1_size/2; i++)
    {
    fft2_float[mm*pa+j  ]=fft1_tmp[fft1_size+2*i  ];
    fft2_float[mm*pa+j+1]=fft1_tmp[fft1_size+2*i+1];
    pa=(pa+1)&timf2_mask;
    }
  }
j=0;    
fft1_nx=(fft1_nx+1)&fft1n_mask;  
fft1_px=(fft1_px+fft1_block)&fft1_mask;  
if( ((fft1_pa-fft1_px+fft1_block+fft1_mask)&fft1_mask) < fft1_block)
  {
  newdat_flag &= -1-ND_FFT1;
  }
timf2_pa=(timf2_pa+fft1_size/2+timf2_mask)&timf2_mask;
newdat_flag|=ND_TIMF2RAW;
}




int powtim_trigger(void)
{
int total_points;
int ia, p0, pa, pb;
float t1,t2;
newdat_flag&=-1-ND_TIMF2RAW;
// timf2_pa is the latest point in timf2 containing valid data. 
total_points=((timf2_pa-timf2_pn1+timf2_mask+1)&timf2_mask)/twice_rxchan;
// Do not run this routine more often than 25 times per second.
if(total_points < ui.input_speed/25)
  {
  return 0;
  }
pa=(timf2_pn1-WIDTH*powtim_xstep+timf2_mask)&timf2_mask;
pb=(timf2_pa-WIDTH*powtim_xstep+timf2_mask)&timf2_mask;
p0=pa;
pa=(pa-WIDTH*powtim_xstep+timf2_mask)&timf2_mask;
t1=0;
ia=p0;
if(sw_onechan)
  {
  while(p0 != pb)
    {
    if(t1 < fft2_float[2*p0])
      {
      t1=fft2_float[2*p0];
      ia=p0;
      }
    p0=(p0+1)&timf2_mask;
    }
  }
else
  {
  while(p0 != pb)
    {
    if(t1 < fft2_float[4*p0]+fft2_float[4*p0+2])
      {
      t1=fft2_float[4*p0]+fft2_float[4*p0+2];
      ia=p0;
      }
    p0=(p0+1)&timf2_mask;
    }
  switch (powtim_trigmode)
    {
    case 0:
// Low rise. Step backwards for 1% power (10% amplitude)
    t2=t1;
    t1*=0.1;
    while(ia != pa && t2 >= t1)
      {
      t2=fft2_float[4*ia]+fft2_float[4*ia+2];
      ia=(ia+timf2_mask)&timf2_mask;
      }
    break;    

    case 1:
// Low fall. Step forwards for 1% power (10% amplitude)
    t2=t1;
    t1*=0.1;
    while(ia != pb && t2 >= t1)
      {
      t2=fft2_float[4*ia]+fft2_float[4*ia+2];
      ia=(ia+1)&timf2_mask;
      }
    break;    

    case 2:
// High rise. Step backwards for 90% power (95% amplitude)
    t2=t1;
    t1*=0.95;
    while(ia != pa && t2 >= t1)
      {
      t2=fft2_float[4*ia]+fft2_float[4*ia+2];
      ia=(ia+timf2_mask)&timf2_mask;
      }
    break;    

    case 3:
// High fall. Step forwards for 90% power (95% amplitude)
    t2=t1;
    t1*=0.95;
    while(ia != pb && t2 >= t1)
      {
      t2=fft2_float[4*ia]+fft2_float[4*ia+2];
      ia=(ia+1)&timf2_mask;
      }
    break;    
    }

  }
timf2_pn2=ia;
timf2_pn1=timf2_pa;
return 1;
}

void powtim_screen(void)
{
char s[80];
unsigned int uj;
int i,j,k,mm, p0;
float t1,f0[4];
int trace_colour[4]={15,12,14,10};
mm=twice_rxchan;
if(timf2_pn2 < 0)return;
// Clear the old traces
if( (powtim_displaymode&4) == 0)
  {
  for(i=0; i<WIDTH; i++)
    {
    for(j=0; j<mm; j++)
      {
      gl_setpixel(i,timf2_pwr_int[mm*i+j],0);
      }
    }
  }  
else
  {
  for(i=1; i<WIDTH; i++)
    {
    for(j=0; j<mm; j++)
      {
      gl_line(i,timf2_pwr_int[mm*i+j],i-1,timf2_pwr_int[mm*(i-1)+j],0);
      }
    }
  powtim_displaymode&=-1-4;  
  }
for(uj=powtim_y0-powtim_scalepix; uj>powtim_y1; uj-=powtim_scalepix)
  {
  gl_hline(0,uj,WIDTH-1,3);
  }
gl_hline(0,powtim_y0,WIDTH-1,7);
for(i=0; i<WIDTH; i+=powtim_scalepix)
  {
  for(uj=powtim_y0; uj<powtim_y0+text_height/2; uj++)
    {
    gl_setpixel(i,uj,15);
    }
  }  
// Display data using timf2_pn2 as pointer.
  
for(j=0; j<mm; j+=2)
  {
  f0[j]=fft2_float[mm*timf2_pn2+j+1];
  sprintf(s,"%f kHz   ",-f0[j]*(0.0005*timf1_sampling_speed/(PI_L)));
  settextcolor(trace_colour[j+1]);
  gl_write(WIDTH/2, powtim_y0+(j/2+1)*text_height, s);
  }
settextcolor(7);
p0=(timf2_pn2-powtim_xstep*WIDTH/4+timf2_mask)&timf2_mask;
if( (powtim_displaymode & 2) ==0)
  {
  for(i=0; i<WIDTH; i++)
    {
    for(j=0; j<mm; j+=2)
      {
      t1=powtim_gain*fft2_float[mm*p0+j];
      if( (powtim_displaymode&1) !=0)
        {
        if(t1<0.0000000001)t1=0.0000000001;
        t1=2*powtim_scalepix*(1+log10(t1));
        }
      t1=powtim_y0-t1;
      if(t1<powtim_y1)t1=powtim_y1;
      if(t1 > powtim_y0)t1=powtim_y0;
      timf2_pwr_int[mm*i+j]=t1;
      gl_setpixel(i,timf2_pwr_int[mm*i+j],trace_colour[j]);
      t1=powtim_fgain*(fft2_float[mm*p0+j+1]-f0[j]);
      t1=powtim_y2-t1;
      if(t1 < powtim_y1)t1=powtim_y1;
      if(t1 > powtim_y0)t1=powtim_y0;
      timf2_pwr_int[mm*i+j+1]=t1;
      gl_setpixel(i,timf2_pwr_int[mm*i+j+1],trace_colour[j+1]);
      }
    p0=(p0+powtim_xstep)&timf2_mask;
    k=(timf2_pa-p0+timf2_mask+1)&timf2_mask;
    if(k >=timf2_mask-powtim_xstep)goto showx; 
    }
  }
else
  {
  powtim_displaymode|=4;
  for(i=0; i<WIDTH; i++)
    {
    for(j=0; j<mm; j+=2)
      {
      t1=powtim_gain*fft2_float[mm*p0+j];
      if( (powtim_displaymode&1) !=0)
        {
        if(t1<0.0000000001)t1=0.0000000001;
        t1=2*powtim_scalepix*(1+log10(t1));
        }
      t1=powtim_y0-t1;
      if(t1<powtim_y1)t1=powtim_y1;
      if(t1 > powtim_y0)t1=powtim_y0;
      timf2_pwr_int[mm*i+j]=t1;
      if(i!=0)gl_line(i-1,timf2_pwr_int[mm*(i-1)+j],i,timf2_pwr_int[mm*i+j],
                                                          trace_colour[j]);
      t1=powtim_fgain*(fft2_float[mm*p0+j+1]-f0[j]);
      t1=powtim_y2-t1;
      if(t1 < powtim_y1)t1=powtim_y1;
      if(t1 > powtim_y0)t1=powtim_y0;
      timf2_pwr_int[mm*i+j+1]=t1;
      if(i!=0)gl_line(i-1,timf2_pwr_int[mm*(i-1)+j+1],i,timf2_pwr_int[mm*i+j+1],
                                                         trace_colour[j+1]);
      }
    p0=(p0+powtim_xstep)&timf2_mask;
    k=(timf2_pa-p0+timf2_mask+1)&timf2_mask;
    if(k >=timf2_mask-powtim_xstep)goto showx; 
    }
  }
showx:;  
}

void powtim_parmwrite(void)
{
char *tmtxt[POWTIM_TRIGMODES]={"Low rise","Low fall","High rise","High fall"};
char *dmtxt[2]={"Lin","Log"};
char s[80];

if(pause_flag == 0)
  {
  xytext(10,0,"Press P for PAUSE to view data in memory.");
  }
else
  {
  clear_lines(0,0);
  settextcolor(14);
  xytext(10,0,"Press R to restart sampling.");
  settextcolor(7);
  }
sprintf(s,"Trig mode: %s  (M to change)  ",tmtxt[powtim_trigmode]);
xytext(20,2,s);
sprintf(s,"Display mode: %s  (L to change)  ",dmtxt[powtim_displaymode&1]);
xytext(20,3,s);
xytext(20,4,"W to toggle between line or pixels");

sprintf(s,"+/-: Amplitude scale %f     ",powtim_gain);
gl_write(5*text_width, powtim_y0+text_height, s);
sprintf(s,"I/D: Time scale %f ms ",
                     powtim_scalepix*1000*powtim_xstep/timf1_sampling_speed);
gl_write(5*text_width, powtim_y0+2*text_height, s);
sprintf(s,"%c/%c: Frequency scale %f kHz ",24,25,
              powtim_scalepix*0.0005*timf1_sampling_speed/(powtim_fgain*PI_L));
gl_write(5*text_width, powtim_y0+3*text_height, s);
}


void power_vs_time(void)
{
int i, j, lirsta;
audio_buf_info ad_info, ad_info2;
float t1;
char s[80];
// This routine stops normal processing and shows RF envelope vs time
// and frequency vs time as an oscilloscope. 
vga_clear();
i=0;
if(genparm[SECOND_FFT_ENABLE] == 0)i=1;
if(disksave_flag != 0)i|=2;
if( fft_cntrl[FFT1_BCKCURMODE].mmx != 0)i|=4;
if(fft_cntrl[FFT2_CURMODE].mmx != 0)i|=8;
if(genparm[SECOND_FFT_SINPOW] == 0)i|=16;
if(genparm[FIRST_FFT_SINPOW] != 2)i|=32;
if(network_flag != 0)i|=64;
if(i != 0)
  {
  settextcolor(12);
  for(j=0; j<8; j++) xytext(j*10,2,"ERROR");
  settextcolor(15);
  xytext(5,4,"Parameters incompatible with RF envelope oscilloscope.");
  settextcolor(7);
  j=7;
  if((i&1)!=0)
    {
    xytext(5,j,"Enable second fft");
    j++;
    }
  if((i&2)!=0)
    {
    xytext(5,j,"Save/read on hard disk not allowed");
    j++;
    }
  if((i&4)!=0)
    {
    xytext(5,j,"No MMX for backwards fft1");
    j++;
    }
  if((i&8)!=0)
    {
    xytext(5,j,"No MMX for fft2");
    j++;
    }
  if((i&16)!=0)
    {
    xytext(5,j,"Window N for fft2 must not be zero");
    j++;
    }
  if((i&32)!=0)
    {
    xytext(5,j,"Window N for fft1 must be two");
    j++;
    }
  if((i&64)!=0)
    {
    xytext(5,j,"Network not allowed");
    j++;
    }
  xytext(9,j+2,"Press any key");
  while(vga_getkey == 0);
  await_keyboard();
  clean_vga_inkey();  
  return;  
  }
j=0;
ensure_valid_input();
ampinfo_flag=1;
lirsta=RX_USER_CONTROL;
timf2_pa=0;
timf2_pn1=0;
timf2_pn2=-1;
newdat_flag&=-1-ND_TIMF2RAW;
// We use timf2_pwr to store screen data.
powtim_y0=0.8*HEIGHT;
powtim_y1=0.1*HEIGHT;
powtim_y2=0.5*HEIGHT;
powtim_scalepix=(powtim_y0-powtim_y1)/10;
powtim_powfac=0.2*HEIGHT/(sqrt((float)(fft1_size))*0x10000);
powtim_gain=1;
powtim_fgain=500;
powtim_xstep=1;
powtim_trigmode=0;
powtim_displaymode=0;
for(i=0; i<WIDTH*2*ui.rx_channels; i++)timf2_pwr_int[i]=powtim_y0;
for(i=0; i<WIDTH; i++)gl_setpixel(i,powtim_y0,15);
// Reconfigure the fft2 tables so they will work for backwards
// fft1.
init_fft(0,fft1_n, fft1_size, fft2_tab, fft2_permute);
timf2_mask = max_fft2n*fft2_size/2-1;
pause_flag=0;
newdat_flag=0;
powtim_parmwrite();
powtim_loop:;
if(pause_flag == 0)
  {
  if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1079);
  if(ui.ad_device_no > 255)
    {
    if(ioctl(audio_in2,SNDCTL_DSP_GETISPACE,&ad_info2)==-1)end_program(1079);
    if(ad_info2.bytes < ad_info.bytes)ad_info=ad_info2;
    ad_info.bytes<<=1;
    }
  if(ad_info.bytes >= ad_read_bytes)
    {
    newdat_flag|=ND_HWARE;
    lirsta=RX_GET_RAW_DATA;
    }
  }  
switch(lirsta)
  {
  case RX_GET_RAW_DATA:
  read_rawdat(0);
  if( (newdat_flag & ND_TIMF1) != 0)
    {
    lirsta=RX_FFT1_FIRST;
    }
  else
    {  
    lirsta=RX_FFT1_SECOND;
    }
  break;

  case RX_FFT1_FIRST:
  if( (fullbuf_flag & ND_FFT1) != 0)
    {
    if( fft1_na != fft1_nb)
      {
      fft1_sumsq_counter=0;
      fft1_c();
      }
    else
      {
      if(lirsta == RX_GET_RAW_DATA)lirsta=RX_FFT1_SECOND;
      }
    }
  else
    {    
    fft1_b(0);
    if( (newdat_flag & ND_TIMF1) == 0)lirsta=RX_FFT1_SECOND;
    }
  break; 

  case RX_FFT1_SECOND:
  fft1_sumsq_counter=-1;
  while( fft1_na != fft1_nb)
    {
    fft1_c();
    }
  lirsta=RX_FFT1_THIRD;
  break;

  case RX_FFT1_THIRD:
  while(fft1_sumsq_pws != fft1_sumsq_pa )
    {
    update_fft1_slowsum();
    }
  lirsta=RX_FFT1_TO_TIMF2;
  break;

  case RX_FFT1_TO_TIMF2:
  make_rf_envelopes();
  if( (newdat_flag & ND_FFT1) == 0)
    {
    lirsta=RX_POWTIM_TRIGGER;
    }
  break;

  case RX_POWTIM_TRIGGER:
  if( (newdat_flag&ND_TIMF2RAW) != 0)
    {
    i=powtim_trigger();
    }
  if(i==0)
    {  
    lirsta=RX_USER_CONTROL;
    }
  else
    {  
    lirsta=RX_POWTIM_SHOW;
    }
    
  break;

  case RX_POWTIM_SHOW:
  powtim_screen();
  lirsta=RX_USER_CONTROL;
  break;

  case RX_USER_CONTROL:
  vga_inkey=vga_getkey();
  if(vga_inkey != 0)
    {
    clean_vga_inkey();
    switch (vga_inkey)
      {
      case 'X':
      return;


      case '+':
      powtim_gain*=2;
      break;
      
      case '-':
      powtim_gain/=1.8;
      break;

      case ARROW_UP_KEY:
      powtim_fgain*=2;
      break;
      
      case ARROW_DOWN_KEY:
      powtim_fgain/=1.8;
      break;

      case 'D':
      if(powtim_xstep < 5)
        {
        powtim_xstep--;
        }
      else
        {
        powtim_xstep/=1.5;
        }
      if(powtim_xstep<1)powtim_xstep=1;     
      break;
      
      case 'I':
      if(powtim_xstep < 5)
        {
        powtim_xstep++;
        }
      else
        {
        powtim_xstep*=2;
        }
      break;

      case 'L':
      powtim_displaymode^=1;
      break;

      case 'W':
      powtim_displaymode^=2;
      break;

      case 'M':
      powtim_trigmode++;
      if(powtim_trigmode >= POWTIM_TRIGMODES)powtim_trigmode=0;
      break;

      case 'P':
      if(pause_flag == 0)
        {
        close_audio_input();
        pause_flag=1;
        }
      break;
        
      case 'R':
      if(pause_flag != 0)
        {
        open_audio_input();
        ensure_valid_input();
        pause_flag=0;
        }
      break;

      case 'Z':
      clear_workload_timers();
      break;

      if(pause_flag != 0)
        {
        case ARROW_RIGHT_KEY:
        timf2_pn2=(timf2_pn2-WIDTH*powtim_xstep/4+timf2_mask+1)&timf2_mask;
        break;

        case ARROW_LEFT_KEY:
        timf2_pn2=(timf2_pn2+WIDTH*powtim_xstep/5)&timf2_mask;
        break;
        }        
      }
    powtim_parmwrite();
    if(pause_flag != 0)
      {
      powtim_screen();
      }
    }  
  else
    {
    if(pause_flag == 0)
      {
      modesub_idle(5);
      t1=0;
      for(i=0; i<ui.ad_channels; i++)
        {
        if(t1<ad_maxamp[i])t1=ad_maxamp[i];
        }
      if(t1<1)t1=1;
      t1/=0x8000;
      t1=-20*log10(t1);
      if(t1 < 0.0000001) t1=0.0000001;
      sprintf(s,"A/D margin %5.2f dB",t1);
      xytext(0,screen_last_line-1,s);
      }
    else
      {
      clear_lines(screen_last_line - 1,screen_last_line);
      }
    }
  break;
  }
goto powtim_loop;
}
