

#include <vga.h>
#include <vgagl.h>
#include <unistd.h>
#include <sys/mman.h>
#include "netdef.h"
#include "conf.h"
#include "globdef.h"
#include "uidef.h"
#include "fft1def.h"
#include "fft2def.h"
#include "fft3def.h"
#include "screendef.h"
#include "seldef.h"
#include "sigdef.h"

#define LIRMOD_INPUT 1
#define LIRMOD_OUTPUT 2
#define LIRMOD_CONTINUE 3
#define LIRMOD_IDLE 4
#define IDLE_ENSURE_MAX 40



#define RX_LOW_PRIO -1
#define RX_GET_RAW_DATA 10
#define RX_FFT1_FIRST 12
#define RX_FFT1_SECOND 14
#define RX_FFT1_THIRD 15
#define RX_FFT1_FOURTH 16
#define RX_FFT1_LAST 19
#define RX_FFT1_TO_TIMF2 25
#define RX_RUN_BLANKER 30
#define RX_CHK_FFT2 40
#define RX_MIX1_BLOCK_FFT1 50
#define RX_MIX1_BLOCK_FFT2 55
#define RX_SHOW_AFC 58
#define RX_FFT3_BLOCK 60
#define RX_SHOW_BASEB_SPECTRUM 65
#define RX_MIX2_BLOCK 70
#define RX_BASEB_DETECT 75
#define RX_AUDIO_BLOCK 80
#define RX_SEND_AUDIO 85
#define RX_WATMO 90
#define RX_USER_CONTROL 100
#define RX_MIX1_BLOCK_TXTEST 300
#define RX_SHOW_FFT1 310
#define RX_SHOW_WIDE 320

#define MAX_USER 3  //Interrupt rate is around 30Hz so this is 0.1 second

// *********************  newdat_flag *************************
// newdat_flag holds info about the amount of unprocessed data.
// A bit that is set indicates that the amount of data is large
// enough for the next processing stage.
// Bits are cleared at the point when the amount of data is smaller
// than required to process one more block.
// ND_HWARE =>    Raw data buffer in device driver. 
//                Input is automatically generated by interrupt driven 
//                routines which may cause overrun errors if data is 
//                not fetched in time. 
//                This flag is set and reset here depending on the result
//                of an ioctl() call to the A/D converter or network.
//                In case data is read from the hard disk this flag is
//                set by a timer in the idle routine until end of file 
//                is encountered. When reading from disk, this flag is
//                reset by read_rawdat.
// ND_TIMF1 =>    This flag is set by read_rawdat if there is enough
//                data for a new fft1 transform. It is reset by fft1_b
// ND_FFT1 =>     This flag is set by fft1_c when at least one new
//                fft1 transform is present. It is reset at different
//                places depending on what parameters the operator has
//                selected. 
// ND_TIMF2RAW => Set by make_timf2, reset by first_noise_blanker.
// ND_TIMF2BLN => Set by first_noise_blanker, reset by make_fft2. 
// ND_FFT2 =>     Set by make_fft2, reset by fft2_mix1_fixed or fft2_mix1_afc
// ND_TIMF3 =>    Set by mix1 routines, reset by make_fft3_all  
// ND_FFT3 =>     set by make_fft3_all, 


//                reset in mix2_loop 
// ND_BASEB =>    a new block of data present in the baseband
//                set by fft3_mix2  
//                reset by make_audio_signal
// ND_DASIG =>    a new block of data present for D/A
//                set by make_audio_signal
//                reset by make_audio_output
// ND_DABUF       Not used
// 
//
//
// *********************  fullbuf_flag *************************
// fullbuf_flag holds info about full buffers.
// A bit that is set indicates that there is not enough space for the 
// next processing stage.
// ND_HWARE =>    set by read_rawdat, reset by fft1_b
// ND_FFT1 =>     set by fft1_b, reset by fft1_c
// ND_TIMF2RAW => set by make_timf2, reset by make_fft2
// ND_FFT2 =>     Set by make_fft2, reset by fft2_mix1_fixed or fft2_mix1_afc
// ND_TIMF3 =>    Set by mix1 routines, reset by make_fft3_all  
// ND_FFT3 =>     set by make_fft3_all, 

// ND_DABUF       Indicates whether the D/A buffer was full when
//                we last looked at it.

void touch_stack(void)
{
int x[4096];
int i;
for(i=0; i<4096; i+=16)
x[i]=0;
}


void wcw_rx(void)
{
char s[80];
int lir_mode, lirsta, input_status, output_status, latest_task;
int i,j,k;
int usercontrol_counter, idle_ensure_counter;
int syncflag, baseb_graph_time;
audio_buf_info ad_info, ad_info2;
touch_stack();
if(ui.memlock == 0)
  {
  if(mlockall(MCL_CURRENT) != 0)end_program(1165);
  }
baseb_graph_time=0;
get_buffers(0);
if(lir_status != LIR_OK)return;
open_audio_input(); 
open_audio_output();
if( genparm[SECOND_FFT_ENABLE] != 0 )init_blanker();
syncflag=0;
// Open the graph windows on the screen.
// Note that they must be opened in this order because
// Each window is placed outside the previous ones and the init
// routines assumes the order below and does not check for all possible
// conflicts.
if(disksave_flag < 2)read_freq_control_data();
init_wide_graph();
if(lir_status != LIR_OK)goto wcw_exit;
if(genparm[SECOND_FFT_ENABLE] != 0)
  {
  init_hires_graph();
  if(lir_status != LIR_OK)goto wcw_exit;
  }
if(genparm[AFC_ENABLE] != 0)
  {
  init_afc_graph();
  if(lir_status != LIR_OK)goto wcw_exit;
  }
init_baseband_graph();
if(lir_status != LIR_OK)goto wcw_exit;
if(ui.rx_channels == 2)
  {
  init_pol_graph();
  }
if(lir_status != LIR_OK)goto wcw_exit;
init_coherent_graph();
if(eme_flag != 0)init_eme_graph();
if(disksave_flag < 2)init_freq_control();
init_users_control_window();
show_name_and_size();
open_mouse();
settextcolor(7);
fflush(NULL);
sync();
clear_da();
ensure_valid_input();
new_baseb_flag=-1;
usercontrol_counter=0;
idle_ensure_counter=0;
clear_workload_timers();
disksync_time=recent_time;
diskread_timer=recent_time;
latest_idle_time=recent_time;
input_status=RX_LOW_PRIO;
output_status=RX_LOW_PRIO;
lirsta=RX_LOW_PRIO;
lir_mode=LIRMOD_IDLE;
i=0;
if(ui.memlock == 0)
  {
  if(mlockall(MCL_CURRENT) != 0)lir_status=LIR_MEMERR;
  }
latest_task=0;
// *******************************************************
//                   MAIN RECEIVE LOOP
// Check if we need to fetch data from the input or if we have to
// write to the output.
// If no immediate action is needed, check lir_mode to decide if there is
// something that was interrupted by input or output and that needs to be
// continued. 
// If there is nothing to continue, check for processing needed at the
// output side as indicated by output_status.
// If output_status is RX_LOW_PRIO, do whatever indicated by input_status.
// *******************************************************
while(lir_status == LIR_OK)
  {
  if(idle_ensure_counter > IDLE_ENSURE_MAX*1.2)goto force_idle;
  flowcnt[27]++;
// find out if there is input data waiting.
  if(disksave_flag < 2)
    {
    if(network_flag == 2)
      {
      if(ioctl(network_fd, FIONREAD, &ad_info.bytes)==-1)end_program(1146);
      ad_info.bytes-=16;
      if(ad_info.bytes<0)ad_info.bytes=0;
      }
    else
      {
      if(ioctl(audio_in,SNDCTL_DSP_GETISPACE,&ad_info)==-1)end_program(1079);
      if(ui.ad_device_no > 255)
        {
        if(ioctl(audio_in2,SNDCTL_DSP_GETISPACE,&ad_info2)==-1)end_program(1079);
        if(ad_info2.bytes < ad_info.bytes)ad_info=ad_info2;
        ad_info.bytes<<=1;
        }
      }
    if(network_flag != 2)
      {
      if(ad_info.fragments >= overrun_limit)
        {
        if(disksave_flag < 2) overrun_flag++;
        }
      if(overrun_flag != 0)
        {
        close_audio_output();
        close_audio_input();
        overrun_count+=overrun_flag;
        overrun_flag=0;
        sprintf(s,"OVERRUN ERRORS %d [%d] ",overrun_count,latest_task);
        wg_error(s,0);
        open_audio_input();
        open_audio_output();    
// Clear disk I/O in case debug messages were causing audio error
        fflush(NULL);
        sync();
        ensure_valid_input();
        clear_da();
        if(new_baseb_flag >= 0)new_baseb_flag=2;
        usercontrol_counter=MAX_USER+1;
        goto user_kbd; 
        }
      }
    if(ad_info.bytes >= ad_read_bytes)
      {
      newdat_flag|=ND_HWARE;
// Go unconditionally to the raw data input routine if we have
// data waiting. Avoiding overrun errors has highest priority.
// Do NOT set lir_mode, we may be doing something that has to 
// finished before processing the input too much.
      input_status=RX_GET_RAW_DATA;
      goto get_raw_data;
      }
    else
      {
      newdat_flag&=-1-ND_HWARE;
      }      
    if(idle_ensure_counter >IDLE_ENSURE_MAX)
      {
force_idle:;
      idle_ensure_counter=0;
      check_timing(0);
      modesub_idle(0);
      latest_task=91;
      if(lir_status != LIR_OK)goto wcw_exit;
      }
// If we arrive here, there is no panic on the input side.
// In case the output is running, do some work at the output side if
// there is room for more data in the output buffer.  
    if(new_baseb_flag == 0)
      {
      if((fullbuf_flag&ND_DABUF) != 0)
        {
        if(ioctl(audio_out,SNDCTL_DSP_GETOSPACE, &da_info) == -1)
                                                         err_restart_da(1108);
        if(da_info.bytes > 2*daout_block ) 
          {
          fullbuf_flag &=-1-ND_DABUF;
          }
        }
      if((fullbuf_flag&ND_DABUF) == 0 && (newdat_flag&ND_DASIG) != 0)
        {
        output_status=RX_SEND_AUDIO;
        goto set_lirsta;
        }
      if( (fullbuf_flag&ND_DASIG) == 0 && (newdat_flag & ND_BASEB) != 0)
        {
        output_status=RX_AUDIO_BLOCK;
        goto set_lirsta;
        }
      if( (newdat_flag&ND_FFT3) != 0 && (fullbuf_flag&ND_BASEB) == 0 )
        {
        output_status=RX_MIX2_BLOCK;        
        goto set_lirsta;
        }
      if( (newdat_flag&ND_TIMF3) != 0 && (fullbuf_flag&ND_FFT3) == 0 )
        {
        output_status=RX_FFT3_BLOCK;        
        goto set_lirsta;
        }
      if(genparm[SECOND_FFT_ENABLE] == 0)
        {   
        if( (newdat_flag&ND_FFT1) != 0 &&
            (fullbuf_flag & ND_TIMF3) == 0 &&
             fft1_nm >= afc_tpts)
          {
          output_status=RX_MIX1_BLOCK_FFT1;        
          goto set_lirsta;
          }
        output_status=RX_LOW_PRIO;        
        goto set_lirsta;
        }   
      if( (newdat_flag&ND_FFT2) != 0 )
        {
        if( (fullbuf_flag & ND_TIMF3) == 0 &&
             fft2_nm >= afc_tpts)
          {
          output_status=RX_MIX1_BLOCK_FFT2;        
          goto set_lirsta;
          }
        fullbuf_flag&=-1-ND_TIMF3;
        if( (newdat_flag & ND_TIMF2BLN) != 0)
          {
          output_status=RX_CHK_FFT2;
          goto set_lirsta;
          }
        output_status=RX_LOW_PRIO;        
        }  
      }
    else
      {
      newdat_flag&=-1-ND_HWARE;
      }      
    }
  else
    {
    if( (newdat_flag&ND_HWARE) != 0)input_status=RX_GET_RAW_DATA;
    if(diskread_pause_flag != 0)input_status=RX_USER_CONTROL;
    }
  if(input_status == RX_GET_RAW_DATA)goto get_raw_data;  
  set_lirsta:;
  if(lir_mode != LIRMOD_CONTINUE)
    {
    if(output_status != RX_LOW_PRIO)
      {
      lirsta=output_status;
      lir_mode=LIRMOD_OUTPUT;
      }
    else
      {  
      lirsta=input_status;    
      lir_mode=LIRMOD_INPUT;
      }
    }  
  flowcnt[0]++;

    
  switch(lirsta)
    {
// ***************************   RX_GET_RAW_DATA   *********************
// Read from the device driver.
// A failure to get here in time will cause an overrun error.
// ********************************************************************* 

    case RX_GET_RAW_DATA:
get_raw_data:;  
    flowcnt[1]++;
// Use up data until there is room for a read in case the buffer is full. 
    while( (fullbuf_flag&ND_HWARE) != 0)fft1_b(0);
    read_rawdat(0);
    idle_ensure_counter++;
    latest_task=1;
// Make sure the user has a chance now and then even on a slow computer
// and even if the parameters are unwisely chosen.
    usercontrol_counter++;
    if(usercontrol_counter>MAX_USER)
      {
      lirsta=RX_USER_CONTROL;
      lir_mode=LIRMOD_CONTINUE;
      }
    else
      {
      lir_mode=LIRMOD_INPUT;
      }
    if( (newdat_flag & ND_TIMF1) != 0)
      {
      input_status=RX_FFT1_FIRST;
      break;
      }
    if( fft1_na != fft1_nb)
      {
      input_status=RX_FFT1_SECOND;
      break;
      }
    if(fft1_sumsq_pws != fft1_sumsq_pa ||
      (workload_percentage < 80 || recent_time-wide_graph_time > 1))
      {
      input_status=RX_FFT1_THIRD;
      break;
      }
    if( genparm[SECOND_FFT_ENABLE] != 0 && 
                                 lir_mode == LIRMOD_INPUT)goto fft1_to_fft2;
    input_status=RX_FFT1_LAST;
    break;
  
    case RX_FFT1_FIRST:
// We can arrive here only if (newdat_flag & ND_TIMF1) != 0) 
    flowcnt[2]++;
    i=0;
    while(i<4096)
      {
      fft1_b(0);
      i+=fft1_size;
      latest_task=3;
      if( (newdat_flag & ND_TIMF1) == 0)
        {
        input_status=RX_FFT1_SECOND;
        break;
        }
      }
    break; 

    case RX_FFT1_SECOND:
    flowcnt[3]++;
    while( fft1_na != fft1_nb)
      {
      fft1_c();
      latest_task=4;
      }
    input_status=RX_FFT1_THIRD;
    break;

    case RX_FFT1_THIRD:
    flowcnt[4]++;
    while(fft1_sumsq_pws != fft1_sumsq_pa )
      {
      update_fft1_slowsum();
      latest_task=5;
      }
    if(workload_percentage < 80 || recent_time-wide_graph_time > 1)
      {
      lirsta=RX_FFT1_FOURTH;
      lir_mode=LIRMOD_CONTINUE;    
      break;
      }
    if( genparm[SECOND_FFT_ENABLE] != 0)goto fft1_to_fft2;
    input_status=RX_FFT1_LAST;
    break;

    case RX_FFT1_FOURTH:
    flowcnt[5]++;
    update_wg_spectrum();
    latest_task=6;
    if(wg_xpix1 == wg_last_xpixel) 
      {
      wide_graph_time=recent_time;
      if( genparm[SECOND_FFT_ENABLE] != 0)goto fft1_to_fft2;
      lirsta=RX_FFT1_LAST;
      }  
    break;
  
    case RX_FFT1_LAST:
    flowcnt[6]++;
    if( genparm[SECOND_FFT_ENABLE] == 0 )
      {
      fft1_waterfall();
      latest_task=7;
      if(new_baseb_flag >= 0)
        {
        if( fft1_nm >= afc_tpts)
          {
          input_status=RX_MIX1_BLOCK_FFT1;
          }
        else
          {
          input_status=RX_WATMO;
          }
        }
      else
        {
        input_status=RX_WATMO;
        newdat_flag&=ND_TIMF1+ND_HWARE;
        fullbuf_flag&=ND_TIMF1+ND_HWARE;
        }
      }
    else  
      {
fft1_to_fft2:;
      if( (newdat_flag & ND_FFT1) != 0)
        {
        input_status=RX_FFT1_TO_TIMF2;
        }
      else
        {
        input_status=RX_RUN_BLANKER;
        }
      }
    lir_mode=LIRMOD_INPUT;
    break;

    case RX_MIX1_BLOCK_FFT1:
// ###########################################################
// Third processing block.
// Convert to baseband and reduce sampling speed.
// ###########################################################  
    flowcnt[7]++;
    if( (fullbuf_flag&ND_TIMF3) != 0)goto unconditional_fft3_block;
    if(genparm[AFC_ENABLE] == 0)
      {
// The AFC is disabled. We just process the signal at the frequency
// selected by the user.
      if( (newdat_flag & ND_FFT1) != 0)
        {
        fft1_mix1_fixed();
        latest_task=8;
        if( (newdat_flag & ND_FFT1) != 0)break;
        }
      }  
    else
      {
// The AFC has to use the information contained in fft1 to find and track
// the selected signal. 
      if( fft1_nb != fft1_nc)
        {
        make_afc();
        latest_task=9;
        if(lir_mode==LIRMOD_INPUT &&  fft1_nb != fft1_nc)break;  
        } 
      if( (newdat_flag & ND_FFT1) != 0)
        {
        fft1_mix1_afc();
        latest_task=10;
        if( (newdat_flag & ND_FFT1) != 0)break;  
        }
      }
    if(lir_mode == LIRMOD_INPUT)
      {
      if( genparm[AFC_ENABLE] != 0 && 
          ag.mode_control != 0 &&
          afc_graph_filled != 0 && 
          recent_time > afc_redraw_time+0.4)
        {
        input_status=RX_SHOW_AFC;
        }
      else
        {
        input_status=RX_FFT3_BLOCK;
        }
      }  
    if(lir_mode == LIRMOD_OUTPUT)output_status=RX_FFT3_BLOCK;
    break;

    case RX_FFT1_TO_TIMF2:
    if( (fullbuf_flag&ND_TIMF2RAW) != 0)goto run_blanker;
    flowcnt[8]++;
    i=0;
fft1_to_timf2:;
    make_timf2();
    i+=fft1_size;
    latest_task=11;  
    if( (newdat_flag & ND_FFT1) == 0)
      {
      input_status=RX_RUN_BLANKER;
      }
    else
      {
      if(i<4096)goto fft1_to_timf2;    
      }
    break;

    case RX_RUN_BLANKER:
run_blanker:;
    flowcnt[9]++;
// Run the noise blanker on timf2_weak
    if( (newdat_flag&ND_TIMF2RAW) != 0)
      {
      first_noise_blanker();
      latest_task=12;
      }
    newdat_flag&=-1-ND_TIMF2RAW;
    input_status=RX_CHK_FFT2;
    break;

    case RX_CHK_FFT2:
// Add the two parts of timf2 so we get a single time function
// and produce the second fft.
    flowcnt[10]++;
    if( (fullbuf_flag & ND_FFT2) != 0) goto skip_fft2;
    if(wg_waterf_ymove_ptr != 0 && 
                         make_fft2_status==FFT2_WATERFALL_LINE) goto watmo;
    if( (newdat_flag & ND_TIMF2BLN) != 0)
      {
      make_fft2();
      latest_task=13;
      if( (newdat_flag & ND_TIMF2BLN) != 0)break;
      }
skip_fft2:;
    if(new_baseb_flag >= 0)
      {
      if( fft2_nm >= afc_tpts)
        {
        input_status=RX_MIX1_BLOCK_FFT2;
        }
      else
        {
        input_status=RX_WATMO;
        }
      }
    else
      {
      newdat_flag&=ND_TIMF1+ND_HWARE+ND_FFT1+ND_TIMF2RAW+ND_TIMF2BLN;
      fullbuf_flag&=ND_TIMF1+ND_HWARE+ND_TIMF2RAW+ND_TIMF2BLN;
      input_status=RX_WATMO;
      }
    break;

    case RX_MIX1_BLOCK_FFT2:
// ###########################################################
// Third processing block.
// Convert to baseband and reduce sampling speed.
// ###########################################################  
    flowcnt[11]++;
    if( (fullbuf_flag&ND_TIMF3) != 0)goto fft3_block;
    if(genparm[AFC_ENABLE] == 0)
      {
// The AFC is disabled. We just process the signal at the frequency
// selected by the user.
// The time function is timf2 on which the noise blanker hopefully
// has improved S/N. 
      if( (newdat_flag & ND_FFT2) != 0)
        {
        fft2_mix1_fixed();
        latest_task=14;
        if( (newdat_flag & ND_FFT2) != 0)break;
        }
      }  
    else
      {
// The AFC has to use the information contained in fft2 to find and track
// the selected signal. 
      if( fft2_na != fft2_nc )
        {
        make_afc();
        latest_task=15;
        if( lir_mode==LIRMOD_INPUT && fft2_na != fft2_nc)break;  
        } 
      if( (newdat_flag & ND_FFT2) != 0)
        {
        fft2_mix1_afc();
        latest_task=16;
        if( (newdat_flag & ND_FFT2) != 0)break;  
        }
      }
    if(lir_mode == LIRMOD_INPUT)
      {
      if( genparm[AFC_ENABLE] != 0 && 
          ag.mode_control != 0 &&
          afc_graph_filled != 0 && 
          recent_time > afc_redraw_time+0.4)
        {
        input_status=RX_SHOW_AFC;
        }
      else
        {
        input_status=RX_FFT3_BLOCK;
        }
      }  
    if(lir_mode == LIRMOD_OUTPUT)output_status=RX_FFT3_BLOCK;
    break;

    case RX_SHOW_AFC:
    flowcnt[12]++;
    show_afc();
    latest_task=17;
    if(recent_time == afc_redraw_time)input_status=RX_FFT3_BLOCK;
    break;

    case RX_FFT3_BLOCK:
fft3_block:;
    flowcnt[13]++;
    if( lir_mode==LIRMOD_INPUT && bg_waterf_ymove_ptr > 0)goto bg_watmo;
unconditional_fft3_block:;
    flowcnt[14]++;
    k=0;
    j=fft1_size/2;
    if(j<4096)j=4096;
    while( (newdat_flag & ND_TIMF3) != 0)
      {
      if( (fullbuf_flag&ND_FFT3) != 0)goto mix2_block;
      make_fft3_all();
      latest_task=18;
      if( lir_mode==LIRMOD_INPUT && bg_waterf_ymove_ptr > 0)goto bg_watmo;
      k+=fft3_size;
      if(k >= j && (newdat_flag & ND_TIMF3) != 0)break;
      }    
    if(lir_mode == LIRMOD_INPUT)
      {
      input_status=RX_MIX2_BLOCK;
      if( (workload_percentage<80 || recent_time-baseb_graph_time>1))
        {
        if(bg_redraw_counter > bg.fft_avgnum/12+1)
          {
          input_status=RX_SHOW_BASEB_SPECTRUM;
          }
        }
      }    
    if(lir_mode == LIRMOD_OUTPUT)output_status=RX_MIX2_BLOCK;
    break;

    case RX_SHOW_BASEB_SPECTRUM:
    flowcnt[15]++;
    show_baseband_spectrum();
    latest_task=19;
    if(bg_xpix1 == bg_last_xpixel) 
      {
      baseb_graph_time=recent_time;
      bg_redraw_counter=0;
      input_status=RX_MIX2_BLOCK;
      }  
    break;


    case RX_MIX2_BLOCK:
mix2_block:;
flowcnt[16]++;
    if( (fullbuf_flag&ND_BASEB) != 0 )
      {
      if((fullbuf_flag&ND_DASIG) ==0)goto audio_block;
      if((fullbuf_flag&ND_DABUF) ==0)goto send_audio;
      if(lir_mode == LIRMOD_INPUT)input_status=RX_WATMO;
      if(lir_mode == LIRMOD_OUTPUT)output_status=RX_LOW_PRIO;
      break;
      }
// Apply the desired baseband filter and get the back transform
// from fft3 to baseb. At the same time sampling speed is reduced again.
    if( (newdat_flag&ND_FFT3) != 0)
      {
      fft3_mix2();
      latest_task=20;
      if( lir_mode == LIRMOD_INPUT && (newdat_flag&ND_FFT3) != 0)break;  
      }
    if( (ui.input_mode&NO_DUPLEX) != 0 && audio_out == -1)
      {
      baseb_wts=4;
      baseb_fx=(baseb_pa-4+baseband_size)&baseband_mask;
      newdat_flag&=-1-ND_BASEB;
      }
    if(lirsta != RX_MIX2_BLOCK)break;
    if(lir_mode == LIRMOD_INPUT)
      {
      if( (newdat_flag&ND_BASEB) !=0 && (fullbuf_flag&ND_DASIG) == 0)
        {
        input_status=RX_AUDIO_BLOCK;
        break;
        }
      if( (newdat_flag&ND_DASIG) != 0 && (fullbuf_flag&ND_DABUF) ==0)
        {
        input_status=RX_SEND_AUDIO;
        break;
        }
      input_status=RX_WATMO;
      break;
      }
    if( (newdat_flag&ND_DASIG) != 0 && (fullbuf_flag&ND_DABUF) ==0)
      {
      output_status=RX_SEND_AUDIO;
      break;
      }
    if( (newdat_flag&ND_BASEB) !=0 && (fullbuf_flag&ND_DASIG) == 0)
      {
      output_status=RX_AUDIO_BLOCK;
      break;
      }
    output_status=RX_LOW_PRIO;
    break;


    case RX_AUDIO_BLOCK:
audio_block:;
    flowcnt[17]++;
// ###########################################################
// Last processing block.
// Convert baseband to desired frequency and send
// the audio to the audio device
// ###########################################################  
    if( (newdat_flag&ND_BASEB)!=0)
      {
      make_audio_signal();
      latest_task=21;
      }
    if(lirsta == RX_AUDIO_BLOCK)
      {  
      if(lir_mode == LIRMOD_INPUT)
        {
        if( (newdat_flag&ND_DASIG) != 0 && (fullbuf_flag&ND_DABUF) ==0)
          {
          input_status=RX_SEND_AUDIO;
          break;
          }
        input_status=RX_WATMO;
        break;
        }
      if( (newdat_flag&ND_DASIG) != 0 && (fullbuf_flag&ND_DABUF) ==0)
        {
        output_status=RX_SEND_AUDIO;
        break;
        }
      output_status=RX_LOW_PRIO;
      }
    break;
  
    case RX_SEND_AUDIO:
send_audio:;
    flowcnt[18]++;
    if(new_baseb_flag == 0)
      {  
      if( (newdat_flag&ND_DASIG) != 0)
        {
        make_audio_output();
        latest_task=22;
        }
      }
    else
      {  
      audio_output_start();
      latest_task=23;
      }
    if(lir_mode == LIRMOD_INPUT)input_status=RX_WATMO;
    if(lir_mode == LIRMOD_OUTPUT)output_status=RX_LOW_PRIO;
    break;

    case RX_WATMO:
    flowcnt[19]++;
    if(wg_waterf_ymove_ptr > 0)
      { 
watmo:;
      flowcnt[20]++;
      move_wg_waterfall(); 
      latest_task=24;
      }
    else
      {
      if(bg_waterf_ymove_ptr > 0)
        {
bg_watmo:;      
         flowcnt[21]++;
        move_bg_waterfall(); 
        latest_task=25;
        }
      else  
        {
        if( (newdat_flag&ND_BASEB_DETECT) != 0)
          {
          input_status=RX_BASEB_DETECT;
          }
        else
          {
          input_status=RX_USER_CONTROL;
          }
        }
      }
    break;

    case RX_BASEB_DETECT:
    coherent();
    flowcnt[22]++;
    input_status=RX_USER_CONTROL;
    break;


    case RX_USER_CONTROL:
//**************************************************************
// Lower priority actions.
//**************************************************************
    flowcnt[23]++;
    check_mouse_actions();
user_kbd:; 
    flowcnt[24]++;
    latest_task=26;
    if(check_keyboard_actions() != 0)goto wcw_exit;
    if(disksave_flag == 8) goto user_kbd;
    if(lir_mode == LIRMOD_INPUT)
      {
      input_status=RX_LOW_PRIO;
      }
    else
      {
      lir_mode=LIRMOD_INPUT;
      }  
    usercontrol_counter=0;  
    break;

    case RX_LOW_PRIO:
// Update usercontrol_counter here. 
// There is no hurry so an extra call to user control 
// now and then will not be a problem.
// This should be unnecessary, but it gives the user access to
// the keyboard even if something went wrong with the input routines
// or the timing routine that tells when it is time to read data
// from the disk.
    usercontrol_counter++;  
    if(usercontrol_counter>MAX_USER)input_status=RX_USER_CONTROL;
    latest_task=27;
    flowcnt[25]++;
    if( check_timing(1) == 0)
      {
      modesub_idle(5);
      latest_task=92;
      idle_ensure_counter=0;
      }
    break;
    }
    flowcnt[26]++;
  }
wcw_exit:;
if( (disksave_flag&(2+4+8)) != 0)
  {
  disksave_flag=4;
  fclose(save_file);
  save_file=NULL;
  }
close_all_except_vga();
}
